/* IP/tcp port redirector. Redirect data from source client to destination target via this
 * host, thus overiding routing path in application layer. 
 * (c) Spiros Ioannou 1999
 * Program tested under sunos 5.x Can support maxchildren connections to the same
 * host simultaneously. forks a new process for every connection.
 *
 * Can support separate regular-expression based filters for client->target and target->client data. 
 * Probably this feature is more useful with HTTP. It's a bit smart so as not to apply regular
 * expressions on non-html data. Feature added in 2010.
 * Regular expressions are read on startup from "toclient.pat" and "totarget.pat" files. 
 * Syntax of these files is "<pattern to search><tab><replacement>". Full lines are searched. 
 * If you want to ignore newlines remove the "REG_NEWLINE" flag from regcomp.
 */


#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <poll.h>

#include <stropts.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <strings.h>
#include <stdlib.h>

#include <regex.h>
#define BIGBUFSIZE 1000000

void childhdl (int);
void connectC (char *, char *);

char buffer[BIGBUFSIZE];
char buff[BIGBUFSIZE];
char newstring[BIGBUFSIZE];

long int lsd, b, r, i, client_sd, target_sd, we_full, we_child, cdown;
socklen_t len;
char targethost[64], targetport[10], localport[32];
struct sockaddr_in sa, sc, sconf;
unsigned char in[2];
struct timeval tval;
struct hostent *chost;
fd_set inout_fds;

char header[6000];
char *hp;
int numofsites = 0;
int mypid = 0;
char *ns = NULL;

int debug=0;
int maxchildren=50;
int manipulate=0;

/*regex globals */
#define MAXPATTERNS 128
#define MAXPATTERNSIZE 1024
int npats_totarget,npats_toclient;

char patterns_totarget[MAXPATTERNS][MAXPATTERNSIZE], /* we apply patterns on write, not on read */
     patterns_toclient[MAXPATTERNS][MAXPATTERNSIZE],
     replacements_totarget[MAXPATTERNS][MAXPATTERNSIZE],
     replacements_toclient[MAXPATTERNS][MAXPATTERNSIZE];


char * strnstr(const char *, const char *, size_t );

FILE *dfp;

/*
 *  * Create a new string with [substr] being replaced by [replacement] in [string]
 *   * Returns the new string, or NULL if out of memory.
 *    * The caller is responsible for freeing this new string.
 *     */
char *str_replace (const char *string, const char *substr,
		   const char *replacement)
{
  char *tok = NULL;
  char *newstr = NULL;

  tok = strstr (string, substr);
  if (tok == NULL)
    return strdup (string);
  newstr =
    malloc (strlen (string) - strlen (substr) + strlen (replacement) + 1);
  if (newstr == NULL)
    return NULL;
  memcpy (newstr, string, tok - string);
  memcpy (newstr + (tok - string), replacement, strlen (replacement));
  memcpy (newstr + (tok - string) + strlen (replacement),
	  tok + strlen (substr),
	  strlen (string) - strlen (substr) - (tok - string));
  memset (newstr + strlen (string) - strlen (substr) + strlen (replacement),
	  0, 1);
  return newstr;
}

/* replace until the end of newline */
char *str_replace2 (const char *string, const char *substr,
		    const char *replacement)
{
  char *tok = NULL;
  char *tok2 = NULL;
  char *subend = NULL;
  char *newstr = NULL;

  tok = strstr (string, substr);

  tok2 = strstr (tok + strlen (substr), "\n");
  subend = tok + strlen (substr);
  if (!tok2)
    tok2 = subend;

  if (tok == NULL)
    return strdup (string);
  newstr =
    malloc (strlen (string) - strlen (substr) - (tok2 - subend) +
	    strlen (replacement) + 1);
  if (newstr == NULL)
    return NULL;

  memcpy (newstr, string, tok - string);
  memcpy (newstr + (tok - string), replacement, strlen (replacement));
  memcpy (newstr + (tok - string) + strlen (replacement),
	  tok + strlen (substr) + (tok2 - subend),
	  strlen (string) - strlen (substr) - (tok2 - subend) - (tok -
								 string));
  memset (newstr + strlen (string) - strlen (substr) - (tok2 - subend) +
	  strlen (replacement), 0, 1);
  return newstr;
}

void usage()
{
    printf ("\nIpRedirectd version 0.4 , Spiros Ioannou 1999-2010\n");
    printf ("Usage: ipredirect [-d[d]] [-m <n>] localport targethost targetport\n");
    printf ("\t-d \tdebug connections\n");
    printf ("\t-dd \tdebug all traffic\n");
    printf ("\t-ddd \tdebug also this software\n");
    printf ("\t-x <n> \tfork at most n children\n");
    printf ("\t-m\tmanipulate traffic using to{client,target}.pat\n");
}

int main (int argc, char *argv[])
{
  int one = 1;
  int opt;
  char *p;
  int ctype=1;

  client_sd = -1;
  target_sd = -1;
  we_full = 0;
  we_child = 0;
  cdown = 1;


  while ((opt = getopt(argc, argv, "dmx:")) != -1) {
    switch (opt) {
    case 'd':
        debug++;
        break;
    case 'm':
        manipulate=1;
        break;
    case 'x':
        maxchildren=atoi(optarg);
        break;
    default: /* '?' */
	usage();
        exit(EXIT_FAILURE);
     }
  }


  if ((argc-optind) != 3 ) {
    usage();
    exit (-1);
  }


  printf ("Maximum servers:%d\n", maxchildren);

  signal (SIGCHLD, childhdl);
  /*signal(SIGCHLD, SIG_IGN); */

  strcpy (localport , argv[optind+0]);
  strcpy (targethost, argv[optind+1]);
  strcpy (targetport, argv[optind+2]);

  if (debug) /*eliminate buffering on stdout*/
      setvbuf(stdout, NULL, _IONBF, 0); 

  if (debug) {
    printf("Local port: %s, Target Host:%s, Target Port:%s\n",localport, targethost, targetport);
  }

  /*read regex patterns*/
  npats_totarget=readpatterns_totarget();	
  npats_toclient=readpatterns_toclient();	

  if (debug) 
    printf("%d patterns to target, %d patterns to client\n",npats_totarget,npats_toclient);


  /* Open  the "listening"  socket */
  if ((lsd = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
    perror ("lsd:socket");
    exit (1);
  }
  setsockopt (lsd, SOL_SOCKET, SO_REUSEADDR, (char *) &one, sizeof (one));
  sa.sin_family = AF_INET;
  sa.sin_port = htons ((u_short) atoi (localport));
  sa.sin_addr.s_addr = htonl (INADDR_ANY);

  if (bind (lsd, (struct sockaddr *) &sa, sizeof (sa)) < 0) {
    perror ("lsd:bind");
    exit (errno);
  }

  if (listen (lsd, 10) < 0) {
    perror ("lsd:listen");
    exit (errno);
  }
  len = sizeof (sa);
  strcpy (buffer, "");

  mypid=getpid();

  printf ("[%d]:daemon ready ...\n",mypid);

  while (1) {
    fflush (stdout);
    tval.tv_sec = 0;
    tval.tv_usec = 0;		/* 50 microseconds */

  startselect:

    FD_ZERO (&inout_fds);	/* Initialize file descriptors for select */

    /* Prepare for input from host A */
    if (we_child && client_sd > 0)
      FD_SET (client_sd, &inout_fds);

    /* Prepare for input from host C */
    if (we_child && target_sd > 0)
      FD_SET (target_sd, &inout_fds);

    /* Prepare to accept a connection to our socket */

    if (!we_child)
      FD_SET (lsd, &inout_fds);

    if (debug>=3)
      printf ("[%d]:before select, we_child:%d\n", getpid(), we_child);

    /* Do a non-timeout select (interupt driven) to minimize load */
    do {
      i = select (64, &inout_fds, (fd_set *) 0, (fd_set *) 0,
		  (struct timeval *) 0 /* &tval */ );

      if ((debug>=3) && (i < 0)) {
	printf ("[%d]:select interrupted:", getpid());
	perror ("select");
      }
    }
    while (i == EINTR || i < 0);

    if (debug>=3)
      printf ("[%d]:after select(=%d),we_child:%d\n", mypid, i, we_child);

    /* We are server, and full,and a new connection */

    if (FD_ISSET (lsd, &inout_fds) && we_full >= maxchildren && !we_child) {
      if ((client_sd = accept (lsd, (struct sockaddr *) &sa, &len)) < 0) {
	perror ("client_sd:000:accept");
      }
      else {
	/* deny connection */

	if (debug) {
	  printf ("[%d]:Rejected new connection -- server full,%d\t", mypid, client_sd);
	  printf ("Family=:%d , port=%d address: %s\n", sa.sin_family,
		  ntohs (sa.sin_port), inet_ntoa (sa.sin_addr));
	}

	strcpy (buffer, "\nServer full, sorry ;-).\n");
	write (client_sd, buffer, sizeof (buffer));
	shutdown (client_sd, 2);
	close (client_sd);
	client_sd = -1;
      }
    }
    /* A new connection, ... */
    else if (FD_ISSET (lsd, &inout_fds) && we_full < maxchildren && !we_child) {
      if (!fork ()) {
	mypid = getpid ();
	if (debug>=3)
	  printf ("[%d]:Child %ld is born\n", mypid);

	we_child = 1;
	we_full = 0;
	if ((client_sd = accept (lsd, (struct sockaddr *) &sa, &len)) < 0) {
	  perror ("child accept client_sd");
	  exit (6);
	}
	if (debug)
	  printf ("[%d]:new connection from address: %s port=%d \n",
		mypid, inet_ntoa (sa.sin_addr), ntohs (sa.sin_port));
	close (lsd);

	connectC (targethost, targetport);
	goto startselect;
      }
      else {
	we_full++;
	if (debug>=3)
	  printf ("[%d]:I have %d children now\n", getpid(),we_full);
      }
    }

    /* transfer data from client to target */
    if (we_child && client_sd > 0 && FD_ISSET (client_sd, &inout_fds)) {
      bzero (buffer, sizeof (buffer));

      if (debug)
	printf ("[%d]:Reading from client...\n",mypid);

      /* read data from client */
      b = read (client_sd, buffer, sizeof (buffer));
      buffer[b] = 0;

      if (debug)
	printf ("[%d]:read %d bytes from client...\n",mypid,b);

      if (debug>=2)
	printf ("client->[%d]:#%s#=%d bytes\n", mypid,buffer, b);

      /* If read failed : client closed connection */
      if (b <= 0) {
	/* Free client socket descriptor for a new connection */

	shutdown (client_sd, 2);
	close (client_sd);

	/* Close target sd also to establish new connection in a fresh manner */
	shutdown (target_sd, 2);
	close (target_sd);
	target_sd = -1;

	if (debug) {
	  printf ("[%d]:Closed connection to  address: %s port=%d \n",
		mypid, inet_ntoa (sa.sin_addr), ntohs (sa.sin_port));
	  printf ("[%d]:pid %d:quiting, client closed connection \n", mypid,getpid ());
	}

	exit (0);
      }
      else { //read ok

	/* write client data to target */
	if (manipulate) {
	  if (debug) 
	    printf("[%d]:manipulating content using totarget.pat\n",mypid);

	  if (strnstr (buffer, "User-Agent: ",5000)) {
	    if (debug) 
	      printf("[%d]:applying totarget.pat\n",mypid);

	    r=regreplace_totarget (buffer, newstring, BIGBUFSIZE,npats_totarget);

	    if (debug) 
	      printf("[%d]:ok\n",mypid);

	    strncpy(buffer,newstring,BIGBUFSIZE);
	    b = strlen (buffer);

	  }
	}


	if (debug>=2) {
	  printf ("\n[%d]->target:#%s#=%d bytes\n", mypid,buffer,b);
	  fflush (stdout);
	}

	if (debug) {
	  printf ("[%d]:Writing to target...\n",mypid);
	}
	
	r = write (target_sd, buffer, b);

	if (debug) {
	  printf ("[%d]:written %d bytes out of %d\n",mypid,r,b);
	}
      }
    }
    /* Transfer data from target to client */
    if (we_child && target_sd > 0 && FD_ISSET (target_sd, &inout_fds) && client_sd > 0) {
      bzero (buffer, sizeof (buffer));

      if (debug)
	printf ("[%d]:Reading from target...", mypid);

      /* read data from target */
      b = read (target_sd, buffer, sizeof (buffer));
      buffer[b] = 0;

      if (debug)
	printf ("read %d bytes\n", b);

      if (debug>=2)
	printf ("\ntarget->[%d]:#%s#=%d bytes\n", mypid, buffer, b);

      if (b <= 0) {
	/* target closed connection */
	shutdown (target_sd, 2);
	close (target_sd);
	target_sd = -1;

	/* Tell A nothing more to hope for */
	shutdown (client_sd, 2);
	close (client_sd);
	client_sd = -1;

	if (debug){
	  printf ("[%d]:Closed connection :   address: %s port=%d \n",
		  mypid, inet_ntoa (sa.sin_addr), ntohs (sa.sin_port));
	  printf ("[%d]:quiting, target host closed connection \n", mypid);
	}

	exit (0);
      }
      else {

	if (debug)
	  printf ("[%d]:Writing to client...\n", mypid);

	if (manipulate) {

	  if (debug) {
	    printf("[%d]:manipulating content using toclient.pat\n",mypid);
	    printf("[%d]:original buffer size:%ld\n",mypid,b);
	  }

	  /* new text header */
	  if (strnstr (buffer, "ype: text/html",2048) && (strnstr (buffer,"200 OK",1024)) )
	    ctype=1;
	  /* new non-text header */
	  else if ((strnstr (buffer, "Content-type:",2048) || strnstr (buffer, "Content-Type:",2048)) 
	      && (strnstr (buffer,"200 OK",1024)) && (!strnstr (buffer, "ype: text/html",2048)) )
	    ctype=0;
	  /* process redirections */
	  else if ((!strnstr (buffer,"200 OK",1024)) && (strnstr (buffer, "HTTP/1",2048)) )
	    ctype=1;
	  else if (ctype==0) /* if it was not text, no more processing */
	    ctype=-1;

	  /* else: more data from binary or text, eitherway the old status remains */
	  /* if it is not a header, the old status remains  for the subsequent connection */

	  /* if it is or was html, process all the contents*/
	  if (ctype==1) {
	    if (debug){
	      printf("\n[%d]:content-type == text/html, processing all data\n",mypid);
	    }

	    strncpy(buff,buffer,b);
	    buff[b]=0;
	    regreplace_toclient (buff, newstring, BIGBUFSIZE,npats_toclient);
	    strncpy(buffer,newstring,BIGBUFSIZE);
	    b = strlen (buffer);
	  }
	  /* else process only headers */
	  else if (ctype==0){
	    char *h0;
	    int newb;

	    if (debug) {
	      printf("\n[%d]:content-type != text/html, processing only headers\n",mypid);
	    }

	    strncpy(header,buffer,5000);
	    hp=strnstr(header,"\r\n\r\n",5000);
	    *hp=0;
	    strncpy(buff,header,5000);
	    regreplace_toclient (buff, newstring, 5000,npats_toclient); /* process only header */
	    /* newstring contains new header now */

	    /* append to newstring the old body */
	    /* (hp-header): header size, b: original buffer size, b-(hp-header): original body size including \r\n\r\n*/

	    newb=strlen(newstring)+b-(hp-header);

	    memcpy(newstring+strlen(newstring),
		buffer + (hp-header), 
		b-(hp-header));

	    b=newb;

	    newstring[b]=0;

	    memcpy(buffer,newstring,b);
	    
	  }
	  else {
	    if (debug)
	      printf("[%d]:NOT processed, buffer size:%ld\n",mypid,b);
	  }

	  if (debug)
	    printf("[%d]:new buffer size:%ld\n",mypid,b);
	}/*manipulate*/

	/* Write data to client */
	r = write (client_sd, &buffer, b);

	if (debug)
	  printf ("[%d]:written %d out of %d bytes\n",mypid,r,b);

	if (debug>=2){
	  printf ("[%d]->client:#%s#=%d bytes\n",mypid,buffer,r);
	  fflush (stdout);
	}

      }
    }

  }
  return 0;
}				/* End of main */

void connectC (char *host, char *port)
{

  /* Open target socket C and connect to target host */

  if ((target_sd = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
    perror ("Can't open target socket");
    exit (-2);
  }
  sc.sin_port = htons ((u_short) atoi (port));

  if (!(chost = gethostbyname (host))) {
    fprintf (stderr, "Host not found (%s)\n",host);
    exit (errno);
  }

  if (debug)
    printf ("target address: %s:%d\n\n", (chost->h_name), ntohs (sc.sin_port));

  sc.sin_family = AF_INET;
  memcpy (&sc.sin_addr, chost->h_addr, chost->h_length);

  if (connect (target_sd, (struct sockaddr *) &sc, sizeof (sc)) < 0) {
    perror ("Can't connect to target");
    exit (-3);
  }
  else if (debug)
    printf ("[%d]:Connected to target\n", mypid);

  cdown = 0;
}

int my_usleep (x)
{
  struct timeval time;

  int r;

  time.tv_sec = x / 1000000;
  time.tv_usec = x % 1000000;
  r = select (0, NULL, NULL, NULL, &time);
  return r;
}

void childhdl (int sig)
{
  int i, r;

  while ((r = waitpid (-1, &i, WNOHANG)) > 0)
    we_full--;

  if (debug>=3)
    printf ("[%d]:Inside childhhdl(%d),child %d died, we_full isnow:%d\n",
	  getpid (), sig, r, we_full);

  signal (SIGCHLD, childhdl);

}


/***REGEX**************************************************************/

int readpatterns_totarget(){
  FILE *conf;
  int i=0,npats=0,r;
  char buff[BIGBUFSIZE];
  char pattern[MAXPATTERNSIZE],replacement[MAXPATTERNSIZE];

  conf=fopen("totarget.pat","r");
  if ( conf == NULL){
    fprintf(stderr,"No target pattern file");
    return(-1);
  }

  while (fgets(buff,MAXPATTERNSIZE,conf)!=NULL ) {
    if (buff[strlen(buff)-1]=='\n')
      buff[strlen(buff)-1]='\0';
    if (buff[0] == '#' ) {
	    continue;
    }
    r=sscanf(buff,"%[^\t]\t%[^\t]",pattern,replacement);
    strncpy(patterns_totarget[i],pattern,MAXPATTERNSIZE-1);
    strncpy(replacements_totarget[i],replacement,MAXPATTERNSIZE-1);
    if (r==2) i++;
  }	

  fclose(conf);
  npats=i;
  return(npats);
}

int readpatterns_toclient(){
  FILE *conf;
  int i=0,npats=0,r;
  char buff[MAXPATTERNSIZE];
  char pattern[MAXPATTERNSIZE],replacement[MAXPATTERNSIZE];

  conf=fopen("toclient.pat","r");
  if ( conf == NULL){
    fprintf(stderr,"No client pattern file");
    return(-1);
  }

  while (fgets(buff,MAXPATTERNSIZE,conf)!=NULL ) {
    if (buff[strlen(buff)-1]=='\n')
      buff[strlen(buff)-1]='\0';
    if (buff[0] == '#' ) {
	    continue;
    }
    r=sscanf(buff,"%[^\t]\t%[^\t]",pattern,replacement);
    strncpy(patterns_toclient[i],pattern,MAXPATTERNSIZE-1);
    strncpy(replacements_toclient[i],replacement,MAXPATTERNSIZE-1);
    if (r==2) i++;
  }	

  fclose(conf);
  npats=i;
  return(npats);
}

int regreplace_totarget (char *buff, char * newstring, int bufsize, int npats)
{
  int i,repls,j,result;
  int ii;
  char *currindex;
  char * b0;
  long int offset;
  regex_t regex;
  regmatch_t matches[MAXPATTERNS];

  strncpy(newstring,buff,bufsize-1);
  repls=0;

  b0=buff;

  for (i=0;i<npats;i++) {
    regcomp(&regex,patterns_totarget[i],REG_ICASE|REG_EXTENDED|REG_NEWLINE);
    offset=0;
    

    while (regexec(&regex,buff, regex.re_nsub+1, matches, 0) != REG_NOMATCH ) {
      repls++;
      bzero(newstring,BIGBUFSIZE-1);

      strncat(newstring,buff,matches[0].rm_so);
      strncat(newstring,replacements_totarget[i],strlen(replacements_totarget[i]));

      currindex=buff+matches[0].rm_eo;

      strncat(newstring,currindex,strlen(currindex)); /*the rest of the string */
      strncpy(buff,newstring,bufsize-1);


      buff=buff + matches[0].rm_so + strlen(replacements_totarget[i]);
      offset=buff-b0;
    } /* while new match on rest*/
    regfree(&regex);
    buff=b0;
    strncpy(newstring,buff,bufsize-1);
  } /* for: next pattern */
  return repls;
}


int regreplace_toclient (char *buff, char * newstring, int bufsize, int npats)
{
  int i,repls,j,result;
  int ii;
  char *currindex;
  char * b0;
  long int offset;
  regex_t regex;
  regmatch_t matches[MAXPATTERNS];

  strncpy(newstring,buff,bufsize-1);
  repls=0;

  b0=buff;

  for (i=0;i<npats;i++) {
    regcomp(&regex,patterns_toclient[i],REG_ICASE|REG_EXTENDED|REG_NEWLINE);
    offset=0;
    

    while (regexec(&regex,buff, regex.re_nsub+1, matches, 0) != REG_NOMATCH ) {
      repls++;
      bzero(newstring,BIGBUFSIZE-1);

      strncat(newstring,buff,matches[0].rm_so);
      strncat(newstring,replacements_toclient[i],strlen(replacements_toclient[i]));

      currindex=buff+matches[0].rm_eo;

      strncat(newstring,currindex,strlen(currindex)); /*the rest of the string */
      strncpy(buff,newstring,bufsize-1);


      buff=buff + matches[0].rm_so + strlen(replacements_toclient[i]);
      offset=buff-b0;
    } /* while new match on rest*/
    regfree(&regex);
    buff=b0;
    strncpy(newstring,buff,bufsize-1);
  } /* for: next pattern */
  return repls;
}

/*
 * Find the first occurrence of find in s, where the search is limited to the
 * first slen characters of s.
 */
char * strnstr(const char *s, const char*find, size_t slen)
{
  char c, sc;
  size_t len;

  if ((c = *find++) != '\0') {
	  len = strlen(find);
	  do {
		  do {
			  if (slen-- < 1 || (sc = *s++) == '\0')
				  return (NULL);
		  } while (sc != c);
		  if (len > slen)
			  return (NULL);
	  } while (strncmp(s, find, len) != 0);
	  s--;
  }
  return ((char *)s);
}

